"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/**
 * Make a {@link Comparator} out of a plain comparison function.
 *
 * @nosideeffects
 * @param {(a: T, b: T) => number} cmp
 *   Comparison function. Should return < 0 if `a < b`, 0 if `a === b`, and > 0
 *   if `a > b`.
 * @returns {Comparator<T>} A comparator for values of type `T`.
 * @template T
 */
function makeComparator(cmp) {
    const thenBy = (cmp2) => makeComparator((a, b) => {
        const ord1 = cmp(a, b);
        if (ord1 === 0) {
            return cmp2(a, b);
        }
        else {
            return ord1;
        }
    });
    const result = cmp.bind({});
    return Object.assign(result, cmp, { thenBy, desc: () => desc(cmp) });
}
exports.makeComparator = makeComparator;
/**
 * Create a {@link Comparator} out of some projection of a `T` to a `U`, where
 * `U` is {@link NaturallyOrderable}.
 *
 * @nosideeffects
 * @param {(x: T) => U} selector Selector to project from `T` to `U`.
 * @returns {Comparator<T>} A comparator for values of type `T`.
 * @template T
 * @template U
 */
function comparing(selector) {
    return makeComparator((a, b) => compare(selector(a), selector(b)));
}
exports.comparing = comparing;
/**
 * Create a {@link Comparator} for objects of type `T` by comparing them on some
 * naturally comparable property.
 *
 * @nosideeffects
 * @param {keyof T} key
 *   A property of `T` that extends `NaturallyOrderable`. The value of the
 *   property will be used as basis for all comparisons in the result
 *   `Comparator`.
 * @returns {Comparator<T>}
 *   A comparator that compares any two instances of `T` using the given
 *   property.
 * @template T
 */
function compareOn(key) {
    return makeComparator((a, b) => compare(a[key], b[key]));
}
exports.compareOn = compareOn;
/**
 * Default comparison function for two values of any {@link NaturallyOrderable}
 * type.
 *
 * @nosideeffects
 * @param {T} a Left hand side of the comparison.
 * @param {T} b Right hand side of the comparison.
 * @returns A number indicating the comparison result:
 *          - A result < 0 implies `a < b`,
 *          - A result of 0 implies`a === b`,
 *          - A result > 0 implies `a > b`
 * @template T
 */
function compare(a, b) {
    if (a < b) {
        return -1;
    }
    else if (a > b) {
        return 1;
    }
    else {
        return 0;
    }
}
exports.compare = compare;
function desc(cmp) {
    return makeComparator((a, b) => {
        return cmp(b, a);
    });
}
